from __future__ import print_function

import argparse
import os
import time
from datetime import timedelta

from torchvision import datasets, transforms
import torch
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler
import torch.distributed as dist
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

RANK = int(os.environ.get('RANK', 0)) 
WORLD_SIZE = int(os.environ.get('WORLD_SIZE', 1))
MASTER_ADDR = os.environ.get('MASTER_ADDR', 'localhost')
MASTER_PORT = os.environ.get('MASTER_PORT', "8888")

def should_distribute():
    return dist.is_available() and WORLD_SIZE > 1

def is_distributed():
    return dist.is_available() and dist.is_initialized()

def load_data(train_batch_size, test_batch_size):
    train_dataset = datasets.FashionMNIST('./data', train=True, 
                       transform=transforms.Compose([
                           transforms.ToTensor(),
                           transforms.Normalize((0.1307,), (0.3081,))
                       ]))
    test_dataset = datasets.FashionMNIST('./data', train=False, 
                        transform=transforms.Compose([
                           transforms.ToTensor(),
                           transforms.Normalize((0.1307,), (0.3081,))
                       ]))

    train_sampler = DistributedSampler(train_dataset, num_replicas=WORLD_SIZE, rank=RANK) if is_distributed() else None
    train_loader = DataLoader(
        dataset=train_dataset,
        batch_size=train_batch_size,
        shuffle=(train_sampler is None),
        sampler=train_sampler,
        )

    test_loader = DataLoader(
        dataset=test_dataset,
        batch_size=test_batch_size,
        shuffle=False,
        )
    return train_loader, test_loader, train_sampler

class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 20, 5, 1)
        self.conv2 = nn.Conv2d(20, 50, 5, 1)
        self.fc1 = nn.Linear(4*4*50, 500)
        self.fc2 = nn.Linear(500, 10)

    def forward(self, x):
        x = F.relu(self.conv1(x))
        x = F.max_pool2d(x, 2, 2)
        x = F.relu(self.conv2(x))
        x = F.max_pool2d(x, 2, 2)
        x = x.view(-1, 4*4*50)
        x = F.relu(self.fc1(x))
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)

def train(args, model, device, train_loader, optimizer, epoch):
    model.train()
    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()
        if batch_idx % args.log_interval == 0:
            print('Train Epoch: {} [{}/{} ({:.0f}%)]\tloss={:.4f}'.format(
                epoch, batch_idx * len(data), len(train_loader.dataset),
                100. * batch_idx / len(train_loader), loss.item()))

def test(model, device, test_loader):
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            test_loss += F.nll_loss(output, target, reduction='sum').item() # sum up batch loss
            pred = output.max(1, keepdim=True)[1] # get the index of the max log-probability
            correct += pred.eq(target.view_as(pred)).sum().item()

    test_loss /= len(test_loader.dataset)
    print('\naccuracy={:.4f}\n'.format(float(correct) / len(test_loader.dataset)))

def main():
    # Training settings
    parser = argparse.ArgumentParser(description='PyTorch MNIST Example')
    parser.add_argument('--batch-size', type=int, default=64, metavar='N',
                        help='input batch size for training (default: 64)')
    parser.add_argument('--test-batch-size', type=int, default=1000, metavar='N',
                        help='input batch size for testing (default: 1000)')
    parser.add_argument('--epochs', type=int, default=10, metavar='N',
                        help='number of epochs to train (default: 10)')
    parser.add_argument('--lr', type=float, default=0.01, metavar='LR',
                        help='learning rate (default: 0.01)')
    parser.add_argument('--momentum', type=float, default=0.5, metavar='M',
                        help='SGD momentum (default: 0.5)')
    parser.add_argument('--no-cuda', action='store_true', default=False,
                        help='disables CUDA training')
    parser.add_argument('--seed', type=int, default=1, metavar='S',
                        help='random seed (default: 1)')
    parser.add_argument('--log-interval', type=int, default=10, metavar='N',
                        help='how many batches to wait before logging training status')
    parser.add_argument('--save-model', action='store_true', default=True,
                        help='For Saving the current Model')
    parser.add_argument('--backend', type=str, help='Distributed backend',
                        choices=[dist.Backend.GLOO, dist.Backend.NCCL, dist.Backend.MPI],
                        default='auto')
    args = parser.parse_args()
    torch.manual_seed(args.seed)

    use_cuda = not args.no_cuda and torch.cuda.is_available()
    device = torch.device("cuda" if use_cuda else "cpu")
    print('Using with {}'.format(device))

    if should_distribute():
        if args.backend == 'auto':
            args.backend = dist.Backend.NCCL if device.type == 'cuda' else dist.Backend.GLOO
        dist.init_process_group(
            backend=args.backend, 
            init_method="env://",
            world_size=WORLD_SIZE,
            rank=RANK,
            # long enough for autoscaler gpu node provision
            timeout=timedelta(minutes=30),
        )
        print('Using distributed PyTorch with {} backend'.format(args.backend))

    train_loader, test_loader, train_sampler = load_data(args.batch_size, args.test_batch_size)
    print("Dataset loaded")

    model = Net().to(device)
    print("Model {} loaded to {}".format(model, device))

    if is_distributed():
        Distributor = nn.parallel.DistributedDataParallel
        model = Distributor(model)

    optimizer = optim.SGD(model.parameters(), lr=args.lr, momentum=args.momentum)

    os.system('nvidia-smi')
    start_time = time.time()

    for epoch in range(1, args.epochs + 1):
        if is_distributed():
            train_sampler.set_epoch(epoch)
        train(args, model, device, train_loader, optimizer, epoch)
        test(model, device, test_loader)
    
    os.system('nvidia-smi')
    print("Time elapsed: {}".format(time.time() - start_time))

    if args.save_model and RANK == 0:
        print("saving model on master host")
        if is_distributed():
            model_state_dict = model.module.state_dict() 
        else: 
            model_state_dict = model.state_dict() 
        torch.save(model_state_dict, "mnist_cnn.pt")
        print("model saved to {}".format("mnist_cnn.pt"))

if __name__ == '__main__':
    main()